#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#if defined(sun)
#include <sys/param.h>
#endif
#include "align.h"

#define MAX_NODE_COUNT 4000000L

typedef struct dp_info{
  int prev_node_no;
  int cost;
} dp_info_t;
  
alignment_t *dp(int dim, alignment_t *alignment)
{
  alignment_t *al_p;
  dp_info_t *info;
  int pos[MAX_DIM], prev_pos[MAX_DIM];
  int i, j, k, l, m, n, p, q;
  int edge_cost, al_len, node_count, is_visited;
  int node_no, prev_node_no;
#ifdef VERBOSE
  struct tms t_start, t_search_start, t_search_end, t_end;
#endif

#ifdef VERBOSE
  times(&t_start);
#endif
  if(dim <= 0){
    return NULL;
  }else if(dim == 1){
    return alignment;
  }else if(dim > MAX_DIM){
    fprintf(stderr, "Too large dimension\n");
    exit(-1);
  }
  node_count = 1;
  for(i = 0; i < dim; i++){
    node_count *= (alignment[i].seq[0].len + 1);
    if(node_count > MAX_NODE_COUNT){
      fprintf(stderr, "Too many nodes\n");
      exit(-1);
    }
  }
  if((info = (dp_info_t *)malloc(sizeof(dp_info_t) * node_count)) == NULL){
    fprintf(stderr, "Can't allocate memory for dp\n");
    exit(-1);
  }
#ifdef VERBOSE
  times(&t_search_start);
#endif
  info[0].cost = 0;
  node_no = 1;
  for(i = 0; i < dim - 1; i++){
    pos[i] = 0;
  }
  pos[dim - 1] = 1;
  while(node_no < node_count){
    is_visited = 0;
    for(i = 1; i < (1 << dim); i++){
      n = 1;
      for(j = 0; j < dim; j++){
	if(i & n){
	  if((prev_pos[j] = pos[j] - 1) < 0){
	    goto next_edge;
	  }
	}else{
	  prev_pos[j] = pos[j];
	}
	n <<= 1;
      }
      prev_node_no = prev_pos[0];
      for(j = 1; j < dim; j++){
	prev_node_no
	  = prev_node_no * (alignment[j].seq[0].len + 1) + prev_pos[j];
      }
      edge_cost = 0;
      for(j = 0; j < dim; j++){
	for(k = j + 1; k < dim; k ++){
	  if((pos[j] != prev_pos[j]) && (pos[k] != prev_pos[k])){
	    for(l = 0; l < alignment[j].dim; l++){
	      for(m = 0; m < alignment[k].dim; m++){
		edge_cost
		  += matching_cost(alignment[j].seq[l].str[prev_pos[j]],
				   alignment[k].seq[m].str[prev_pos[k]]);
	      }
	    }
	  }else if(pos[j] != prev_pos[j]){
	    n = 0; 
	    for(l = 0; l < alignment[j].dim; l++){
	      if(is_normal_char(alignment[j].seq[l].str[prev_pos[j]])){
		n++;
	      }
	    }
	    for(m = 0; m < alignment[k].dim; m++){
	      if(((prev_pos[k] == 0)
		  || (alignment[k].seq[m].str[prev_pos[k] - 1]
		      == EXTERNAL_GAP_CHAR))
		 || ((prev_pos[k] == alignment[k].seq[0].len)
		     || (alignment[k].seq[m].str[prev_pos[k]]
			 == EXTERNAL_GAP_CHAR))){
		edge_cost += n * (-external_gap_penalty);
	      }else{
		edge_cost += n * (-internal_gap_penalty);
	      }
	    }
	  }else if(pos[k] != prev_pos[k]){
	    n = 0;
	    for(m = 0; m < alignment[k].dim; m++){
	      if(is_normal_char(alignment[k].seq[m].str[prev_pos[k]])){
		n++;
	      }
	    }
	    for(l = 0; l < alignment[j].dim; l++){
	      if(((prev_pos[j] == 0)
		  || (alignment[j].seq[l].str[prev_pos[j] - 1]
		      == EXTERNAL_GAP_CHAR))
		 || ((prev_pos[j] == alignment[j].seq[0].len)
		     || (alignment[j].seq[l].str[prev_pos[j]]
			 == EXTERNAL_GAP_CHAR))){
		edge_cost += n * (-external_gap_penalty);
	      }else{
		edge_cost += n * (-internal_gap_penalty);
	      }
	    }
	  }
	}
      }
      if((!is_visited)
	 || (info[prev_node_no].cost + edge_cost < info[node_no].cost)){
	is_visited = 1;
	info[node_no].cost = info[prev_node_no].cost + edge_cost;
	info[node_no].prev_node_no = prev_node_no;
      }
    next_edge:
    }
#ifdef VERBOSE
    if((++node_no & 0x7fff) == 0){
      fprintf(stderr,
	      ".... visited %d nodes (forward search)\n", node_no);
    }
#endif
    for(i = dim - 1; i >= 0; i--){
      if(++pos[i] > alignment[i].seq[0].len){
	pos[i] = 0;
      }else{
	break;
      }
    }
  }
#ifdef VERBOSE
  times(&t_search_end);
#endif
  if((al_p = (alignment_t *)malloc(sizeof(alignment_t))) == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  al_p->dim = 0;
  for(i = 0; i < dim; i++){
    al_p->dim += alignment[i].dim;
  }
  al_len = 0;
  for(node_no = node_count - 1; node_no > 0;
      node_no = info[node_no].prev_node_no){
    al_len++;
  }
  if((al_p->seq = (sequence_t *)malloc(sizeof(sequence_t) * al_p->dim))
     == NULL){
    fprintf(stderr, "Can't allocate memory for result alignment");
    exit(-1);
  }
  k = 0;
  for(i = 0; i < dim; i++){
    for(j = 0; j < alignment[i].dim; j++){
      al_p->seq[k].len = al_len;
      if((al_p->seq[k].str
	  = (char *)malloc(sizeof(char) * (al_len + 1))) == NULL){
	fprintf(stderr, "Can't allocate memory for result alignment");
	exit(-1);
      }
      al_p->seq[k].str[al_len] = '\0';
      p = al_len - 1;
      q = alignment[i].seq[0].len - 1;
      for(node_no = node_count - 1; node_no > 0;
	  node_no = info[node_no].prev_node_no){
	n = node_no - info[node_no].prev_node_no;
	for(l = dim - 1; l >= 0; l--){
	  prev_pos[l] = n % (alignment[l].seq[0].len + 1);
	  n /= (alignment[l].seq[0].len + 1);
	}
	if(prev_pos[i] > 0){
	  al_p->seq[k].str[p--] = alignment[i].seq[j].str[q--];
	}else{
	  if(((q + 1 > alignment[i].seq[0].len - 1)
	      || (alignment[i].seq[j].str[q + 1] == EXTERNAL_GAP_CHAR))
	     || ((q < 0)
		 || (alignment[i].seq[j].str[q] == EXTERNAL_GAP_CHAR))){
	    al_p->seq[k].str[p--] = EXTERNAL_GAP_CHAR;
	  }else{
	    al_p->seq[k].str[p--] = INTERNAL_GAP_CHAR;
	  }
	}
      }
      k++;
    }
  }
  al_p->score = -info[node_count - 1].cost;
  for(i = 0; i < dim; i++){
    for(j = 0; j < alignment[i].dim; j++){
      for(k = j + 1; k < alignment[i].dim; k++){
	for(l = 0; l < alignment[i].seq[0].len; l++){
	  al_p->score
	    += -matching_cost(alignment[i].seq[j].str[l],
			      alignment[i].seq[k].str[l]);
	}
      }
    }
  }
  free(info);
#ifdef VERBOSE
  times(&t_end);
  fprintf(stderr, "# of whole nodes: %d\n", node_count);
  fprintf(stderr, "# of visited nodes: %d\n", node_count);
  fprintf(stderr, "Used time:     User    System  Total\n");
#if defined(linux)
  fprintf(stderr, "  Pre-process  %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_start.tms_utime - t_start.tms_utime) / CLK_TCK,
	  (float)(t_search_start.tms_stime - t_start.tms_stime) / CLK_TCK,
	  (float)((t_search_start.tms_utime - t_start.tms_utime)
		  + (t_search_start.tms_stime - t_start.tms_stime)) / CLK_TCK);
  fprintf(stderr, "  Search       %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_end.tms_utime - t_search_start.tms_utime) / CLK_TCK,
	  (float)(t_search_end.tms_stime - t_search_start.tms_stime) / CLK_TCK,
	  (float)((t_search_end.tms_utime - t_search_start.tms_utime)
		  + (t_search_end.tms_stime - t_search_start.tms_stime))
	  / CLK_TCK);
  fprintf(stderr, "  Post-process %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_search_end.tms_utime) / CLK_TCK,
	  (float)(t_end.tms_stime - t_search_end.tms_stime) / CLK_TCK,
	  (float)((t_end.tms_utime - t_search_end.tms_utime)
		  + (t_end.tms_stime - t_search_end.tms_stime)) / CLK_TCK);
  fprintf(stderr, "  Total        %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_start.tms_utime) / CLK_TCK,
	  (float)(t_end.tms_stime - t_start.tms_stime) / CLK_TCK,
	  (float)((t_end.tms_utime - t_start.tms_utime)
		  + (t_end.tms_stime - t_start.tms_stime)) / CLK_TCK);
#elif defined(sun)
  fprintf(stderr, "  Pre-process  %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_start.tms_utime - t_start.tms_utime) / HZ,
	  (float)(t_search_start.tms_stime - t_start.tms_stime) / HZ,
	  (float)((t_search_start.tms_utime - t_start.tms_utime)
		  + (t_search_start.tms_stime - t_start.tms_stime)) / HZ);
  fprintf(stderr, "  Search       %6.2f  %6.2f  %6.2f\n",
	  (float)(t_search_end.tms_utime - t_search_start.tms_utime) / HZ,
	  (float)(t_search_end.tms_stime - t_search_start.tms_stime) / HZ,
	  (float)((t_search_end.tms_utime - t_search_start.tms_utime)
		  + (t_search_end.tms_stime - t_search_start.tms_stime)) / HZ);
  fprintf(stderr, "  Post-process %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_search_end.tms_utime) / HZ,
	  (float)(t_end.tms_stime - t_search_end.tms_stime) / HZ,
	  (float)((t_end.tms_utime - t_search_end.tms_utime)
		  + (t_end.tms_stime - t_search_end.tms_stime)) / HZ);
  fprintf(stderr, "  Total        %6.2f  %6.2f  %6.2f\n",
	  (float)(t_end.tms_utime - t_start.tms_utime) / HZ,
	  (float)(t_end.tms_stime - t_start.tms_stime) / HZ,
	  (float)((t_end.tms_utime - t_start.tms_utime)
		  + (t_end.tms_stime - t_start.tms_stime)) / HZ);
#else
  fprintf(stderr, "  Pre-process  %6d  %6d  %6d\n",
	  (int)(t_search_start.tms_utime - t_start.tms_utime),
	  (int)(t_search_start.tms_stime - t_start.tms_stime),
	  (int)((t_search_start.tms_utime - t_start.tms_utime)
		+ (t_search_start.tms_stime - t_start.tms_stime)));
  fprintf(stderr, "  Search       %6d  %6d  %6d\n",
	  (int)(t_search_end.tms_utime - t_search_start.tms_utime),
	  (int)(t_search_end.tms_stime - t_search_start.tms_stime),
	  (int)((t_search_end.tms_utime - t_search_start.tms_utime)
		+ (t_search_end.tms_stime - t_search_start.tms_stime)));
  fprintf(stderr, "  Post-process %6d  %6d  %6d\n",
	  (int)(t_end.tms_utime - t_search_end.tms_utime),
	  (int)(t_end.tms_stime - t_search_end.tms_stime),
	  (int)((t_end.tms_utime - t_search_end.tms_utime)
		+ (t_end.tms_stime - t_search_end.tms_stime)));
  fprintf(stderr, "  Total        %6d  %6d  %6d\n",
	  (int)(t_end.tms_utime - t_start.tms_utime),
	  (int)(t_end.tms_stime - t_start.tms_stime),
	  (int)((t_end.tms_utime - t_start.tms_utime)
		+ (t_end.tms_stime - t_start.tms_stime)));
#endif
#endif
  return al_p;
}
