#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#if defined(sun)
#include <sys/param.h>
#endif
#include "align.h"

#define h_e_parent(n) ((n) / 2)
#define h_e_left(n) ((n) * 2)
#define h_e_right(n) ((n) * 2 + 1)

void construct_heap(node_t node, node_set_t node_set, int dim, alignment_t *alignment, 
		    int wk, int fs_k, int bs_k, int dir, int delta,
#ifdef VERBOSE
		    int *constructed, int *h_t, int *h_edge,
#endif
		    int add_cost, estimator_t *estimator, int status)
/* Not implemented Eppstein for bidirectional A* */
{
  int h_e_size;
  int i, j, k, l, m, n, dlt, sub;
  int edge_cost, pair_edge_cost, pair_est;
  int previous_pos[MAX_DIM], *pos;
  node_t previous_node;
  edge_heap_t he;
  epp_node_t pht, ht;


/* if the heap has already built, there's nothing to do */
  if (is_constructed(node, dir)) return;

  set_status(node, dir, status);
#ifdef VERBOSE
  (constructed[dir])++;
  if (!(constructed[dir] & 255)) fprintf(stderr, "Constructed for %d nodes...\n",constructed[dir]);
#endif

/* At first, we must construct heap for the previous nodes! */
  construct_heap(lookup_prev_node(node, dir), node_set, dim, alignment, 
		 wk, fs_k, bs_k, dir, delta,
#ifdef VERBOSE
		 constructed, h_t, h_edge,
#endif
		 add_cost, estimator, status);

/* Then, we must make H_edge for this node. */
  /* First, we compute delta */
  h_e_size = 0;

  if ((malloc_edge_heap(node, dir, ((1 << dim)-2))) == NULL) {
    fprintf(stderr,"Cannot allocate memory for H_edge\n");
    exit(-1);
  }

  pos = node->pos;
  if (dir) {  /* BACKWARD */
    for(i = 1; i < (1 << dim); i++){
      n = 1;
      for(j = 0; j < dim; j++){
	if(i & n){
	  if((previous_pos[j] = pos[j] + 1) > alignment[j].seq[0].len){
	    goto bs_another_edge;
	  }
	}else{
	  previous_pos[j] = pos[j];
	}
	n <<= 1;
      }
      if (((previous_node = get_node(node_set, previous_pos)) == NULL)
             || (is_not_finished(previous_node, dir))
	     || (lookup_prev_node(node, dir) == previous_node)) {
	goto bs_another_edge;
      }

      edge_cost = 0;
      sub = 0;

      for(j = 0; j < dim; j++){
	for(k = j + 1; k < dim; k ++){
	  pair_edge_cost = 0;
	  if((previous_pos[j] != pos[j]) && (previous_pos[k] != pos[k])){
	    for(l = 0; l < alignment[j].dim; l++){
	      for(m = 0; m < alignment[k].dim; m++){
		pair_edge_cost += weights[j][k] * wk
		  * matching_cost(alignment[j].seq[l].str[pos[j]],
				  alignment[k].seq[m].str[pos[k]]);
	      }
	    }
	    pair_edge_cost += weights[j][k] * wk * add_cost * 2;
	  }else if(previous_pos[j] != pos[j]){
	    n = 0; 
	    for(l = 0; l < alignment[j].dim; l++){
	      if(is_normal_char(alignment[j].seq[l].str[pos[j]])){
		n++;
	      }
	    }
	    for(m = 0; m < alignment[k].dim; m++){
	      if(((pos[k] == 0)
		  || (alignment[k].seq[m].str[pos[k] - 1]
		      == EXTERNAL_GAP_CHAR))
		 || ((pos[k] == alignment[k].seq[0].len)
		     || (alignment[k].seq[m].str[pos[k]]
			 == EXTERNAL_GAP_CHAR))){
		pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
	      }else{
		pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
	      }
	    }
	    pair_edge_cost += weights[j][k] * wk * add_cost;
	  }else if(previous_pos[k] != pos[k]){
	    n = 0;
	    for(m = 0; m < alignment[k].dim; m++){
	      if(is_normal_char(alignment[k].seq[m].str[pos[k]])){
		n++;
	      }
	    }
	    for(l = 0; l < alignment[j].dim; l++){
	      if(((pos[j] == 0)
		  || (alignment[j].seq[l].str[pos[j] - 1]
		      == EXTERNAL_GAP_CHAR))
		 || ((pos[j] == alignment[j].seq[0].len)
		     || (alignment[j].seq[l].str[pos[j]]
			 == EXTERNAL_GAP_CHAR))){
		pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
	      }else{
		pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
	      }
	    }
	    pair_edge_cost += weights[j][k] * wk * add_cost;
	  }
	  pair_edge_cost
	    -= weights[j][k] * bs_k *
	                (lookup_estimate(estimator[sub],
					 previous_pos[j], previous_pos[k])
			 + add_cost * (previous_pos[j] + previous_pos[k]));
	  pair_est
	    = lookup_estimate(estimator[sub], pos[j], pos[k])
	      + add_cost * (pos[j] + pos[k]);
	  edge_cost += pair_edge_cost + weights[j][k] * bs_k * pair_est;
	  sub++;
	}
      }

      dlt = lookup_cost(previous_node, dir) + edge_cost - lookup_cost(node, dir);
      if (dlt > delta*wk) goto bs_another_edge;

      set_delta_of_edge_heap(node, dir, h_e_size, dlt);
      set_node_of_edge_heap(node, dir, h_e_size, previous_node);
      h_e_size++;

    bs_another_edge:
    }
  } else {  /* FORWARD */
    for(i = 1; i < (1 << dim); i++){
      n = 1;
      for(j = 0; j < dim; j++){
	if(i & n){
	  if((previous_pos[j] = pos[j] - 1) < 0){
	    goto fs_another_edge;
	  }
	}else{
	  previous_pos[j] = pos[j];
	}
	n <<= 1;
      }
      if (((previous_node = get_node(node_set, previous_pos)) == NULL)
             || (is_not_finished(previous_node, dir))
	     || (lookup_prev_node(node, dir) == previous_node)) {
	goto fs_another_edge;
      }

      edge_cost = 0;
      sub = 0;

      for(j = 0; j < dim; j++){
	for(k = j + 1; k < dim; k ++){
	  pair_edge_cost = 0;
	  if((previous_pos[j] != pos[j]) && (previous_pos[k] != pos[k])){
	    for(l = 0; l < alignment[j].dim; l++){
	      for(m = 0; m < alignment[k].dim; m++){
		pair_edge_cost += weights[j][k] * wk
		  * matching_cost(alignment[j].seq[l].str[previous_pos[j]],
				  alignment[k].seq[m].str[previous_pos[k]]);
	      }
	    }
	    pair_edge_cost += weights[j][k] * wk * add_cost * 2;
	  }else if(previous_pos[j] != pos[j]){
	    n = 0; 
	    for(l = 0; l < alignment[j].dim; l++){
	      if(is_normal_char(alignment[j].seq[l].str[previous_pos[j]])){
		n++;
	      }
	    }
	    for(m = 0; m < alignment[k].dim; m++){
		if(((previous_pos[k] == 0)
		    || (alignment[k].seq[m].str[previous_pos[k] - 1]
			== EXTERNAL_GAP_CHAR))
		   || ((previous_pos[k] == alignment[k].seq[0].len)
		       || (alignment[k].seq[m].str[previous_pos[k]]
			   == EXTERNAL_GAP_CHAR))){
		  pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
		}else{
		  pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
		}
	      }
	    pair_edge_cost += weights[j][k] * wk * add_cost;
	  }else if(previous_pos[k] != pos[k]){
	    n = 0;
	    for(m = 0; m < alignment[k].dim; m++){
	      if(is_normal_char(alignment[k].seq[m].str[previous_pos[k]])){
		n++;
	      }
	    }
	    for(l = 0; l < alignment[j].dim; l++){
	      if(((previous_pos[j] == 0)
		  || (alignment[j].seq[l].str[previous_pos[j] - 1]
		      == EXTERNAL_GAP_CHAR))
		 || ((previous_pos[j] == alignment[j].seq[0].len)
		     || (alignment[j].seq[l].str[previous_pos[j]]
			 == EXTERNAL_GAP_CHAR))){
		pair_edge_cost += weights[j][k] * wk * n * (-external_gap_penalty);
	      }else{
		pair_edge_cost += weights[j][k] * wk * n * (-internal_gap_penalty);
	      }
	    }
	    pair_edge_cost += weights[j][k] * wk * add_cost;
	  }
	  pair_edge_cost
	    -= weights[j][k] * fs_k *
	              (lookup_estimate(estimator[sub],
				       previous_pos[j], previous_pos[k])
		       + add_cost * ((alignment[j].seq[0].len
				      - previous_pos[j])
				     + (alignment[k].seq[0].len
					- previous_pos[k])));
	  pair_est
	    = lookup_estimate(estimator[sub], pos[j], pos[k])
	      + add_cost * ((alignment[j].seq[0].len - pos[j])
			    + (alignment[k].seq[0].len - pos[k]));
	  edge_cost += pair_edge_cost + weights[j][k] * fs_k * pair_est;
	  sub++;
	}
      }

      dlt = lookup_cost(previous_node, dir) + edge_cost - lookup_cost(node, dir);
      if (dlt > delta*wk) goto fs_another_edge;

      set_delta_of_edge_heap(node, dir, h_e_size, dlt);
      set_node_of_edge_heap(node, dir, h_e_size, previous_node);
      h_e_size++;

    fs_another_edge:
    }
  }
  if ((realloc_edge_heap(node, dir, h_e_size) == NULL) && (h_e_size != 0)) {
    fprintf(stderr,"Cannot re-allocate memory for H_edge\n");
    exit(-1);
  }
  set_edge_heap_size(node, dir, h_e_size);
#ifdef VERBOSE
  (h_edge[dir]) += h_e_size;
#endif

  /* Make it to 2-heap except for top which has only one child */

  if (h_e_size>=2) {
    he = lookup_edge_heap(node, dir);
    for(i = h_e_parent(h_e_size-1);i > 0;i--) {
      tune_h_edge_down(he, i, h_e_size);
    }
    if (he[0].delta > he[1].delta) {
      swap_h_edge(he, 0, 1);
      tune_h_edge_down(he, 1, h_e_size);
    }
  }

/* At last, we append the top of the H_edge to the copy of previous H_T */

  previous_node = lookup_prev_node(node, dir);
  pht=lookup_tree_heap(previous_node, dir);
  if (h_e_size == 0) {
    set_tree_heap(node, dir, pht);
  } else if (pht == NULL) {
    set_tree_heap(node, dir, (ht = init_epp_node()));
    ht->node = node;
    ht->left = NULL;
    ht->right = NULL;
  } else {
    construct_tree_heap(
#ifdef VERBOSE
			h_t,
#endif
			node, previous_node, dir
			);
  }
}

void tune_h_edge_down(edge_heap_t he, int i, int sz)
/* It must not used in case of i=0 */
{
  int l, r;

  l = h_e_left(i);
  if (l >= sz) return;
  r = l+1;
  if (r>=sz) {
    if (he[i].delta > he[l].delta) swap_h_edge(he,i,l);
    return;
  }
  if (he[l].delta > he[r].delta) {
    if (he[i].delta > he[r].delta) {
      swap_h_edge(he, i, r);
      tune_h_edge_down(he,r,sz);
    }
  } else {
    if (he[i].delta > he[l].delta) {
      swap_h_edge(he,i,l);
      tune_h_edge_down(he,l,sz);
    }
  }
}

void swap_h_edge(edge_heap_t he, int i, int j)
{
  node_t tmp_node;
  int tmp;

  tmp_node = he[i].node;
  he[i].node = he[j].node;
  he[j].node = tmp_node;
  tmp = he[i].delta;
  he[i].delta = he[j].delta;
  he[j].delta = tmp;
}

void construct_tree_heap(
#ifdef VERBOSE
			 int *h_t,
#endif
			 node_t node, node_t prenode, int dir)
{
  epp_node_t pre, new, old, tmp;

  old = lookup_tree_heap(prenode, dir);
  set_tree_heap(node, dir, (new = init_epp_node()));
#ifdef VERBOSE
  (h_t[dir])++;
#endif
  pre = NULL;
  new->left = NULL; /* At first, we use it as parent ! */
  while(old != NULL){
    new->node = old->node;
    new->right = old->left;  /* to make balanced ! */
    pre = new;
    new = init_epp_node();
#ifdef VERBOSE
    (h_t[dir])++;
#endif
    new->left = pre; /* parent */
    old = old->right;
  }
  new->node = node;
  new->left = NULL;
  new->right = NULL;

  while(pre != NULL) {
    if (lookup_ht_delta(new, dir) < lookup_ht_delta(pre, dir)) {
      swap_epp_node(new, pre);
    } else {
      break;
    }
    tmp = pre;
    pre = pre->left; /* parent */
    tmp->left = new; /* left */
    new = tmp;
  }
  while(pre != NULL) {
    tmp = pre;
    pre = pre->left; /* parent */
    tmp->left = new; /* left */
    new = tmp;
  }
}

epp_node_t init_epp_node(void)
{
  epp_node_t en;

  if ((en = (epp_node_t)malloc(sizeof(struct epp_node))) == NULL) {
    fprintf(stderr,"Can't allocate memory for H_T\n");
    exit(-1);
  }
  return en;
}

void swap_epp_node(epp_node_t p, epp_node_t q)
{
  node_t node;

  node = p->node;
  p->node = q->node;
  q->node = node;
}
